open Syntax

let rec lookup (env:env) (var:string) : value = match env with
  | [] -> failwith "variable not found"
  | b::env2 -> if var = b.b_name then b.b_value else lookup env2 var

let extract_num (v:value) : int = match v with
  | Num(n) -> n
  | _ -> failwith "expected Num"

let extract_closure (v:value) : closure = match v with
  | Closure(c) -> c
  | _ -> failwith "expected Closure"

let rec eval (env:env) (e:expr) : value = match e with
  | Var x -> lookup env x
  | Value v -> v
  | Plus (e1, e2) -> Num((extract_num (eval env e1)) + extract_num (eval env e2))
  | Let l -> let v = eval env l.local in eval ({b_name = l.let_var; b_value = v}::env) l.let_body
  | Fn f -> Closure { env = env; fn = f}
  | Call call -> let c = extract_closure(eval env call.callee) in
                 let a = eval env call.arg in
                 let formal = c.fn.formal in
                 eval ({b_name = formal.tb_name; b_value = a}::c.env) c.fn.fn_body

let rec t_lookup (env:t_env) (var:string) : typ = match env with
  | [] -> failwith "variable not found"
  | b::env2 -> if var = b.tb_name then b.typ else t_lookup env2 var

let check_type t1 t2 : unit = if t1 = t2 then () else failwith "type mismatch"

let typecheck_val (v:value) : typ = match v with
  | Num _ -> Number
  | Closure _ -> failwith "shouldn't be closures at typechecking time"

let rec typecheck (env:t_env) (e:expr) : typ = match e with
  | Var x -> t_lookup env x
  | Value v -> typecheck_val v
  | Plus (e1, e2) -> check_type (typecheck env e1) Number; check_type (typecheck env e2) Number; Number
  | Let l -> let tau1 = typecheck env l.local in typecheck ({tb_name = l.let_var; typ = tau1} :: env) l.let_body
  | Fn f -> typecheck (f.formal :: env) f.fn_body
  | Call c -> 
          let callee_type = typecheck env c.callee in match callee_type with
            | Func arrow_type ->
                      let arg_type = typecheck env c.arg in begin
                        check_type arrow_type.arg_type arg_type;
                        arrow_type.result_type
                      end
            | _ -> failwith "arrow type expected at a call"

