(* Type *)
type typ =
  | Func of fn_typ
  | Number
[@@deriving show { with_path = false }]

and fn_typ =
  { arg_type : typ;
    result_type : typ;
  }
[@@deriving show { with_path = false }]

(* Expression *)
and expr =
  | Var of string
  | Value of value
  | Plus of expr * expr
  | Let of lt
  | Fn of fn
  | Call of call
[@@deriving show { with_path = false }]

and value =
  | Num of int
  | Closure of closure
[@@deriving show { with_path = false }]

and lt =
  { let_var : string;
    local : expr;
    let_body : expr;
  }
[@@deriving show { with_path = false }]

and fn =
  { formal : typ_binding;
    fn_body : expr;
  }
[@@deriving show { with_path = false }]

and typ_binding =
  { tb_name : string;
    typ : typ;
  }

and binding =
  { b_name : string;
    b_value : value;
  }

and closure =
  { env : env;
    fn : fn;
  }

(*  CallExpression *)
and call =
  { callee : expr;
    arg: expr;
  }
[@@deriving show { with_path = false }]

and env = binding list

and t_env = typ_binding list

